def _colored_by_color_code(text: str, color_code: int, bold: bool = False, underline: bool = False):
	"""
		Returns an ANSI colored string by its color code,
		and can also be bold and underlined.
	"""
	options = "\033["
	if bold:
		options += "1;"
	if underline:
		options += "4;"
	options += f"{color_code}m"
	return f"{options}{text}\033[0m"



def colored(text: str, color: str, bold: bool = False, underline: bool = False):
	"""
		Returns an ANSI colored string which can be bold and underlined.
		Allowed colors are: black, red, green, orange, blue, magenta,
		cyan, white, bright_black, bright_red, bright_green, bright_orange,
		bright_blue, bright_magenta, bright_cyan, bright_white.
	"""
	colors = {
		"black": 30,
		"red": 31,
		"green": 32,
		"orange": 33,
		"blue": 34,
		"magenta": 35,
		"cyan": 36,
		"white": 37,
		"bright_black": 90,
		"bright_red": 91,
		"bright_green": 92,
		"bright_orange": 93,
		"bright_blue": 94,
		"bright_magenta": 95,
		"bright_cyan": 96,
		"bright_white": 97,
	}
	if color not in colors:
		raise Exception(f"Color {color} does not exist")
	return _colored_by_color_code(text, colors[color], bold=bold, underline=underline)


import time
import logging
import functools
from datetime import datetime

from superpy import string



class _ColoredLoggingFormatter(logging.Formatter):
	def __init__(self, ts_color: str = None):
		self.ts_color = ts_color
		super().__init__(fmt="$levelno: $msg", datefmt="%Y.%m.%d %H:%M:%S", style="$")


	def format(self, record):
		base = string.colored("[$name]\t$asctime\t", self.ts_color)
		formats = {
			logging.CRITICAL: f"{base}{string.colored('$msg', 'bright_magenta')}",
			logging.ERROR: f"{base}{string.colored('$msg', 'bright_red')}",
			logging.WARNING: f"{base}{string.colored('$msg', 'bright_orange')}",
			logging.INFO: f"{base}{string.colored('$msg', 'bright_white')}",
			logging.DEBUG: f"{base}{string.colored('$msg', 'white')}",
		}
		formatter = logging.Formatter(
			fmt=formats.get(record.levelno, self._fmt),
			datefmt="%Y-%m-%d %H:%M:%S",
			style="$",
		)
		return formatter.format(record)



def make_logger(name: str, ts_color: str = "bright_orange", terminal: bool = True, files=[], level=logging.DEBUG):
	"""
		Create a styled logger with nice formating that writes to
		the terminal and/or a set of files.
	"""
	logger = logging.getLogger(name)
	logger.setLevel(level)
	if terminal:
		console_handler = logging.StreamHandler()
		console_handler.setFormatter(_ColoredLoggingFormatter(ts_color=ts_color))
		logger.addHandler(console_handler)
	for file in files:
		file_handler = logging.FileHandler(file)
		file_handler.setFormatter(_ColoredLoggingFormatter(ts_color=ts_color))
		logger.addHandler(file_handler)
	return logger



class log:
	"""
		decorator for logging.
		filepath: write to file
		with_args: list of indices of args that should apppear in the log
		with_kwargs: list of kwarg names that should appear in the log
	"""
	def __init__(self, filepath=None, with_args=[], with_kwargs=[], color="white"):
		# there is no need to make a class for a decorator if there are no parameters
		self.filepath = filepath
		self.with_args = with_args
		self.with_kwargs = with_kwargs
		self.color = color

	def __call__(self, method):
		# returns the decorator itself, which accepts a function and returns another function
		# wraps ensures that the name and docstring of 'method' is preserved in 'wrapper'
		@functools.wraps(method)
		def wrapper(*args, **kwargs):
			# the wrapper passes all parameters to the function being decorated
			timestamp = datetime.now().strftime("%Y.%m.%d %H:%M:%S")
			t1 = time.time()
			res = method(*args, **kwargs)
			t2 = time.time()

			ms = f"{((t2 - t1) * 1000):.1f}ms"
			if len(ms) < 10:
				ms = (" " * (10 - len(ms))) + ms

			log = f"{timestamp}{ms}    {method.__name__}"
			arglist = [f"{a}" for i, a in enumerate(args) if i in self.with_args]
			kwarglist = [f"{k}={a}" for k, a in kwargs.items() if k in self.with_kwargs]
			log += "(" + ", ".join(arglist + kwarglist) + ")"

			if self.filepath is None:
				print(log)
			else:
				with open(self.filepath, "a") as f:
					f.write(f"\n{log}")
			return res
		return wrapper


import os
import time
import random

from superpy import log



def _generate_bytes(size_bytes: int, times: int = 1):
	"""
		Generates a list of {times} random bytes objects,
		each the size of {size_bytes}.
	"""
	return [os.urandom(size_bytes) for _ in range(times)]



def _write_bytes(bytes_to_write_list: list, file_names: list):
	"""
		Write each item of the {bytes_to_write_list} to its own file,
		specified in {file_names}.
		The writing duration is measured, and prined at the end.
	"""
	sec = 0.0
	for file_name, bytes_to_write in zip(file_names, bytes_to_write_list):
		start_time = time.time()
		with open(file_name, "wb") as f:
			f.write(bytes_to_write)
		end_time = time.time()
		sec += end_time - start_time

	mb = (len(bytes_to_write) * len(file_names)) / 1_000_000.0
	print(f"Write {mb:.2f}mb over {len(file_names)} files took {1000 * sec:.1f}ms at {mb / sec:.2f}mb/s")



def _read_bytes(file_names: list):
	"""
		Read each file specified in {file_names}.
		The reading duration is measured, and prined at the end.
	"""
	sec, size = 0.0, 0
	for file_name in file_names:
		start_time = time.time()
		with open(file_name, "rb") as f:
			read_bytes = f.read()
		end_time = time.time()
		sec += end_time - start_time
		size += len(read_bytes)

	mb = size / 1_000_000.0
	print(f"Read {mb:.2f}mb over {len(file_names)} files took {1000 * sec:.1f}ms at {mb / sec:.2f}mb/s")



@log(with_args=[0, 1])
def test_write_read_speed(size_bytes: int, times: int = 1):
	"""
		Generates a list of {times} random bytes objects,
		each the size of {size_bytes}, and then writes and reads them
		sequentially, while printing the results.
	"""
	try:
		os.makedirs(".sp_test_write_read_speed/")
		file_names = [f".sp_test_write_read_speed/{random.randint(0, 99999999999999)}" for _ in range(times)]
		bytes_to_write_list = _generate_bytes(size_bytes, times=times)
		_write_bytes(bytes_to_write_list, file_names)
		_read_bytes(file_names)
	finally:
		for file_name in file_names:
			os.remove(file_name)
		os.rmdir(".sp_test_write_read_speed/")


from concurrent.futures import ThreadPoolExecutor, wait

def run_threaded(fns_args: [(callable, tuple)], max_threads=None):
	"""
		Run a list of tasks concurrently, and return their results as
		a list in the same order. A task is a 2-tuple of the function and an
		n-tuple of the function's n arguments.
		Remember: A 1-tuple needs a trailing comma, eg. (x,)
	"""
	if max_threads is None:
		max_threads = len(fns_args)
	results = []
	with ThreadPoolExecutor(max_threads) as pool:
		for fn, args in fns_args:
			future = pool.submit(fn, *args)
			results.append(future)
	wait(results)
	return [r.result() for r in results]


def deep_update(base_dict: dict, update_dict: dict):
	"""
		Works like update, but recursively on each level
	"""
	res = base_dict.copy()
	for key, val in update_dict.items():
		c = isinstance(res.get(key), dict) and isinstance(val, dict)
		res[key] = deep_update(res[key], val) if c else val
	return res


import traceback
import time

from superpy import string



class test:
	"""
		A decorator that runs tests automatically when running a file.
		Provide a setup function, and it will run before the test.
		Provide a teardown function, and it will run after the test,
		even if an exception occurs.
	"""
	def __init__(self, setup: callable = lambda: None, teardown: callable = lambda: None):
		self.setup = setup
		self.teardown = teardown


	def __call__(self, method):
		self.setup()
		try:
			t1 = time.time()
			method()
			t2 = time.time()
			ms = f"{((t2 - t1) * 1000):.1f}ms"
			print(string.colored(f"Test {method.__name__}() finished in {ms}", "bright_green"))
		except AssertionError:
			print(string.colored(f"Test {method.__name__}() failed!", "bright_red"))
			print(string.colored(traceback.format_exc(), "bright_red"))
		except BaseException:
			raise
		finally:
			self.teardown()



class test_class:
	"""
		A decorator that runs all tests of a class automatically whose function names
		start with "test_".
		The class_setup runs before all tests, and the class_teardown runs after all tests.
		each_setup and each_teardown runs before and after each test.
	"""
	def __init__(self, class_setup: callable = lambda: None, class_teardown: callable = lambda: None, each_setup: callable = lambda: None, each_teardown: callable = lambda: None):
		self.class_setup = class_setup
		self.class_teardown = class_teardown
		self.each_setup = each_setup
		self.each_teardown = each_teardown


	def do_test(self, method):
		self.each_setup()
		try:
			t1 = time.time()
			method()
			t2 = time.time()
			ms = f"{((t2 - t1) * 1000):.1f}ms"
			print("--> " + string.colored(f"Test {method.__name__}() finished in {ms}", "bright_green"))
		except AssertionError:
			print("--> " + string.colored(f"Test {method.__name__}() failed!", "bright_red"))
			print(string.colored(traceback.format_exc(), "bright_red"))
		except BaseException:
			self.class_teardown()
			raise
		finally:
			self.each_teardown()


	def __call__(self, method):
		print(f"Run test class {method.__name__}:")
		self.class_setup()
		obj = method()
		for test_method_name in [m for m in dir(obj) if m.startswith("test_")]:
			self.do_test(getattr(obj, test_method_name))
		self.class_teardown()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['super_py']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'super-py',
    'version': '0.1.0',
    'description': 'Features that Python should have in the standard library',
    'long_description': '# SuperPy\n\nDescription follows (Or look into the files, its not that complicated)',
    'author': 'Marcel Kröker',
    'author_email': 'kroeker.marcel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# flake8: noqa
from superpy.logging import make_logger, log
from superpy.testing import test, test_class
from superpy import disk
from superpy import string
from superpy import concurrency
from superpy import dicts

################################################################################
######## How to make decorators ################################################
################################################################################

# # No arguments decorator: @decorator
# def decorator(function):
# 	# wraps ensures that the name and docstring of 'function' is preserved in 'wrapper'
# 	@functools.wraps(function)
# 	def wrapper(*args, **kwargs):
# 		return function(*args, **kwargs)
# 	return wrapper

# # Decorator with arguments: @decorator(arg1, arg2=1)
# class decorator:
# 	def __init__(self, arg1, arg2=1):
# 		# there is no need to make a class for a decorator if there are no parameters
# 		self.arg1 = arg1
# 		self.arg2 = arg2

# 	def __call__(self, function):
# 	# returns the decorator itself, which accepts a function and returns another function
# 	# wraps ensures that the name and docstring of 'function' is preserved in 'wrapper'
# 		@functools.wraps(function)
# 		def wrapper(*args, **kwargs):
# 			return function(*args, **kwargs)
# 		return wrapper
